program ELASWM2D

  implicit none

  integer :: n1,n2,nt,nts,i1,i2,it
  integer :: i1s,i2s,cpt,cpt_snap
  real :: h,dt
  real,dimension(:),allocatable :: source_tmp,source
  real,dimension(:,:),allocatable :: vp,vs,rho,eps,delta
  real,dimension(:,:),allocatable :: c11,c33,c44,c13,b
  real,dimension(:,:),allocatable :: ux,uz,sigma_xx,sigma_zz,sigma_xz
  character(len=20),external :: str
  
  !Spong layers
  integer :: k,width
  real :: val
  real,dimension(:),allocatable :: spg_x,spg_z
  real,dimension(:,:),allocatable :: spg

  
  !####################################################
  ! INPUT PARAMETERS
  !####################################################
  
  !Hard coded simulation parameters 
  n1=1501
  n2=1501
  nts=2000
  nt=3000
  h=2.5e0
  dt=0.0005


  write(*,*) '====================='
  write(*,*) ' PARAMETER SETTINGS  '
  write(*,*) '====================='
  write(*,*) 'n1 : ',n1
  write(*,*) 'n2 : ',n2
  write(*,*) 'nts : ',nts  
  write(*,*) 'nt : ',nt
  write(*,*) 'h : ',h
  write(*,*) 'dt : ',dt

  
  !Hard coded medium properties
  allocate(vp(n1,n2),vs(n1,n2),rho(n1,n2),eps(n1,n2),delta(n1,n2))
  vp(:,:)=2000e0
  vs(:,:)=1000e0
  rho(:,:)=2000e0
  eps(:,:)=0.0e0
  delta(:,:)=0.0e0
  !eps(:,:)=0.3e0
  !delta(:,:)=0.1e0

  !CFL TEST
  write(*,*) 'CFL IS : ',h/maxval(vp(:,:))/sqrt(2e0)
  write(*,*) 'dt IS : ',dt
  
  if(dt> (h/maxval(vp(:,:))/sqrt(2e0))) then
     write(*,*) 'WARNING: CFL CONDITION NOT SATISFIED'
  endif

  !####################################################
  ! SOURCE 
  !####################################################

  !Time signature read from file
  allocate(source_tmp(nts))
  open(10,file='source',access='direct',recl=4*nts)
  read(10,rec=1) source_tmp(:)
  close(10)
  allocate(source(nt))
  source(:)=0e0
  do it=1,nts
     source(it)=source_tmp(it)
  enddo
  
  !Spatial location on one grid point
  i1s=n1/2
  i2s=n2/2
  write(*,*) 'The source will be located at point : ',i1s,i2s
    
  !###############################################################
  ! INITIALIZE SCHEME
  !###############################################################

  !Cij coefficients depending on vp,vs,rho,epsilon and delta
  allocate(c11(n1,n2),c13(n1,n2),c33(n1,n2),c44(n1,n2))
  c11(:,:)=rho(:,:)*(vp(:,:)**2)*(1e0+2e0*eps(:,:))
  c33(:,:)=rho(:,:)*(vp(:,:)**2)
  c44(:,:)=rho(:,:)*(vs(:,:)**2)
  c13(:,:)=-c44(:,:)+sqrt(rho(:,:)**2*(vp(:,:)**2-vs(:,:)**2)**2+2e0*delta(:,:)*(rho(:,:)**2)*(vp(:,:)**2)*(vp(:,:)**2-vs(:,:)**2))  

  !Buoyancy
  allocate(b(n1,n2))
  b(:,:)=1e0/rho(:,:)
  
  !###############################################################
  ! CHECK CONSISTENT CIJ
  !###############################################################
  open(10,file='c11',access='direct',recl=4*n1*n2)
  write(10,rec=1) c11(:,:)
  close(10)
  open(10,file='c33',access='direct',recl=4*n1*n2)
  write(10,rec=1) c33(:,:)
  close(10)
  open(10,file='c44',access='direct',recl=4*n1*n2)
  write(10,rec=1) c44(:,:)
  close(10)
  open(10,file='c13',access='direct',recl=4*n1*n2)
  write(10,rec=1) c13(:,:)
  close(10)
  open(10,file='b',access='direct',recl=4*n1*n2)
  write(10,rec=1) b(:,:)
  close(10)

  !###############################################################
  ! ALLOCATE/INITIALIZE SPONGE LAYERS 
  !###############################################################
  allocate(spg_z(n1))
  allocate(spg_x(n2))
  spg_z(:)=1e0
  spg_x(:)=1e0
  width=200
  do k=1,width
     val=exp(-(0.8*(width-k)/width)**2)
     spg_z(k)=val
     spg_z(n1-k+1)=val
     spg_x(k)=val
     spg_x(n2-k+1)=val
  enddo
  allocate(spg(n1,n2))
  do i2=1,n2
     do i1=1,n1
        spg(i1,i2)=spg_z(i1)*spg_x(i2)
     enddo
  enddo
  open(10,file='spg',access='direct',recl=4*n1*n2)
  write(10,rec=1) spg(:,:)
  close(10)
  !BYPASS SPONGE LAYERS
  spg(:,:)=1e0

  
  !###############################################################
  ! ALLOCATE/INTIIALIZE FIELDS 
  !###############################################################
  allocate(ux(n1,n2),uz(n1,n2),sigma_xx(n1,n2),sigma_zz(n1,n2),sigma_xz(n1,n2))
  ux(:,:)=0e0
  uz(:,:)=0e0
  sigma_xx(:,:)=0e0
  sigma_zz(:,:)=0e0
  sigma_xz(:,:)=0e0

 
  !###############################################################
  ! MAIN LOOP
  !###############################################################
  cpt=0
  cpt_snap=0
  do it=1,nt

     !Staggered grid scheme
     do i2=2,n2-1
        do i1=2,n1-1
           ux(i1,i2)=ux(i1,i2)+b(i1,i2)*dt/h*(sigma_xx(i1,i2+1)-sigma_xx(i1,i2))+b(i1,i2)*dt/h*(sigma_xz(i1+1,i2)-sigma_xz(i1,i2))
           uz(i1,i2)=uz(i1,i2)+b(i1,i2)*dt/h*(sigma_zz(i1+1,i2)-sigma_zz(i1,i2))+b(i1,i2)*dt/h*(sigma_xz(i1,i2+1)-sigma_xz(i1,i2))
           
           sigma_xx(i1,i2)=sigma_xx(i1,i2)+c11(i1,i2)*dt/h*(ux(i1,i2)-ux(i1,i2-1))+c13(i1,i2)*dt/h*(uz(i1,i2)-uz(i1-1,i2))
           sigma_zz(i1,i2)=sigma_zz(i1,i2)+c13(i1,i2)*dt/h*(ux(i1,i2)-ux(i1,i2-1))+c33(i1,i2)*dt/h*(uz(i1,i2)-uz(i1-1,i2))
           
           sigma_xz(i1,i2)=sigma_xz(i1,i2)+c44(i1,i2)*dt/h*(ux(i1,i2)-ux(i1-1,i2)+uz(i1,i2)-uz(i1,i2-1))           
        enddo
     enddo

     !Source
     uz(i1s,i2s)=uz(i1s,i2s)+dt*source(it)          
     !ux(i1s,i2s)=ux(i1s,i2s)+dt*source(it)
     !sigma_xx(i1s,i2s)=sigma_xx(i1s,i2s)+0.5*dt*source(it)     
     !sigma_zz(i1s,i2s)=sigma_zz(i1s,i2s)+0.5*dt*source(it)

     !Sponge layer
     ux(:,:)=ux(:,:)*spg(:,:)
     uz(:,:)=uz(:,:)*spg(:,:)
     sigma_xx(:,:)=sigma_xx(:,:)*spg(:,:)
     sigma_zz(:,:)=sigma_zz(:,:)*spg(:,:)
     sigma_xz(:,:)=sigma_xz(:,:)*spg(:,:)

     !Snapshot output
     cpt=cpt+1
     !if(cpt.eq.5) then  => use this value for having 1 snapshot each 5 time-iterations : VIDEOS settings
     if(cpt.eq.200) then !=> use this value for having 1 snapshot each 200 time-iterations : IMAGE settings
        cpt=0
        write(*,*) ' ITERATION ', it
        open(10,file='snap_uz'//str(cpt_snap),access='direct',recl=4*n1*n2)
        write(10,rec=1) uz(:,:)
        close(10)
        open(10,file='snap_ux'//str(cpt_snap),access='direct',recl=4*n1*n2)
        write(10,rec=1) ux(:,:)
        close(10)        
        cpt_snap=cpt_snap+1
     endif
  enddo
   
end program ELASWM2D


! Function to convert an integer to string
character(len=20) function str(k)
  implicit none
  integer, intent(in) :: k
  write (str, *) k
  str = adjustl(str)
end function str
